from __future__ import annotations

from itertools import count
from typing import Iterable

from vsexprtools import ExprOp, ExprVars, combine, norm_expr
from vstools import CustomIntEnum, CustomNotImplementedError, FuncExceptT, PlanesT, StrList, flatten_vnodes, vs

__all__ = [
    'MeanMode'
]


class MeanMode(CustomIntEnum):
    MINIMUM = -2
    HARMONIC = -1
    GEOMETRIC = 0

    ARITHMETIC = 1

    RMS = 2
    CUBIC = 3
    MAXIMUM = 4

    LEHMER = 10

    MINIMUM_ABS = 20
    MAXIMUM_ABS = 21

    MEDIAN = 30

    def __call__(
        self, *_clips: vs.VideoNode | Iterable[vs.VideoNode], planes: PlanesT = None, func: FuncExceptT | None = None
    ) -> vs.VideoNode:
        func = func or self.__class__

        clips = flatten_vnodes(_clips)

        n_clips = len(clips)
        n_op = n_clips - 1

        if n_clips < 2:
            return next(iter(clips))

        if self == MeanMode.MINIMUM:
            return ExprOp.MIN(clips, planes=planes, func=func)

        if self == MeanMode.MAXIMUM:
            return ExprOp.MAX(clips, planes=planes, func=func)

        if self == MeanMode.GEOMETRIC:
            return combine(clips, ExprOp.MUL, None, None, [1 / n_clips, ExprOp.POW], planes=planes, func=func)

        if self == MeanMode.LEHMER:
            counts = range(n_clips)
            clip_vars = ExprVars(n_clips)

            expr = StrList([[f'{clip} neutral - D{i}!' for i, clip in zip(counts, clip_vars)]])

            for y in range(2):
                expr.extend([
                    [f'D{i}@ {3 - y} pow' for i in counts],
                    ExprOp.ADD * n_op, f'P{y + 1}!'
                ])

            expr.append('P2@ 0 = 0 P1@ P2@ / ? neutral +')

            return norm_expr(clips, expr, planes=planes, func=func)

        if self in {MeanMode.RMS, MeanMode.ARITHMETIC, MeanMode.CUBIC, MeanMode.HARMONIC}:
            return combine(
                clips, ExprOp.ADD, f'{self.value} {ExprOp.POW}', None, [
                    n_clips, ExprOp.DIV, 1 / self, ExprOp.POW
                ], planes=planes, func=func
            )

        if self in {MeanMode.MINIMUM_ABS, MeanMode.MAXIMUM_ABS}:
            operator = ExprOp.MIN if self is MeanMode.MINIMUM_ABS else ExprOp.MAX

            expr_string = ''
            for src in ExprVars(n_clips):
                expr_string += f'{src} neutral - abs {src.upper()}D! '

            for i, src, srcn in zip(count(), ExprVars(n_clips), ExprVars(1, n_clips)):
                expr_string += f'{src.upper()}D@ {srcn.upper()}D@ {operator} {src} '

                if i == n_clips - 2:
                    expr_string += f'{srcn} '

            expr_string += '? ' * n_op

            return norm_expr(clips, expr_string, planes=planes, func=func)

        if self == MeanMode.MEDIAN:
            all_clips = str(ExprVars(1, n_clips))

            n_ops = n_clips - 2

            yzmin, yzmax = [
                all_clips + f' {op}' * n_ops for op in (ExprOp.MIN, ExprOp.MAX)
            ]

            return norm_expr(
                clips, f'{yzmin} YZMIN! {yzmax} YZMAX! x YZMIN@ min x = YZMIN@ x YZMAX@ max x = YZMAX@ x ? ?', planes
            )

        raise CustomNotImplementedError



from __future__ import annotations

from typing import Any

from vsexprtools import norm_expr
from vstools import (
    CustomTypeError, PlanesT, VSFunction, check_ref_clip, check_variable, FunctionUtil, normalize_planes, vs, ConvMode
)

from .blur import gauss_blur, min_blur, box_blur, median_blur
from .enum import BlurMatrix
from .limit import limit_filter
from .rgtools import repair
from .util import normalize_radius

__all__ = [
    'unsharpen',
    'unsharp_masked',
    'limit_usm',
    'fine_sharp',
    'soothe'
]


def unsharpen(
    clip: vs.VideoNode, strength: float = 1.0, sigma: float | list[float] = 1.5,
    prefilter: vs.VideoNode | VSFunction | None = None, **kwargs: Any
) -> vs.VideoNode:
    assert check_variable(clip, unsharpen)

    ref = prefilter(clip) if callable(prefilter) else prefilter

    check_ref_clip(clip, ref)

    den = ref or clip
    blur = gauss_blur(den, sigma, **kwargs)

    unsharp = norm_expr([den, blur], f'x y - {strength} * x +', 0)

    if ref is not None:
        return unsharp

    return unsharp.std.MergeDiff(clip.std.MakeDiff(den))


def unsharp_masked(
    clip: vs.VideoNode, radius: int | list[int] = 1, strength: float = 100.0, planes: PlanesT = None
) -> vs.VideoNode:
    planes = normalize_planes(clip, planes)

    if isinstance(radius, list):
        return normalize_radius(clip, unsharp_masked, radius, planes, strength=strength)

    blurred = BlurMatrix.LOG(radius, strength=strength)(clip, planes)

    return norm_expr([clip, blurred], 'x dup y - +')


def limit_usm(
    clip: vs.VideoNode, blur: int | vs.VideoNode | VSFunction = 1, thr: int | tuple[int, int] = 3,
    elast: float = 4.0, bright_thr: int | None = None, planes: PlanesT = None
) -> vs.VideoNode:
    """Limited unsharp_masked."""

    if callable(blur):
        blurred = blur(clip)
    elif isinstance(blur, vs.VideoNode):
        blurred = blur
    elif blur <= 0:
        blurred = min_blur(clip, -blur, planes=planes)
    elif blur == 1:
        blurred = BlurMatrix.BINOMIAL()(clip, planes)
    elif blur == 2:
        blurred = BlurMatrix.MEAN()(clip, planes)
    else:
        raise CustomTypeError("'blur' must be an int, clip or a blurring function!", limit_usm, blur)

    sharp = norm_expr([clip, blurred], 'x dup y - +', planes)

    return limit_filter(sharp, clip, thr=thr, elast=elast, bright_thr=bright_thr)


def fine_sharp(
        clip: vs.VideoNode, mode: int = 1, sstr: float = 2.0, cstr: float | None = None, xstr: float = 0.19,
        lstr: float = 1.49, pstr: float = 1.272, ldmp: float | None = None, planes: PlanesT = 0
) -> vs.VideoNode:
    from scipy import interpolate

    func = FunctionUtil(clip, fine_sharp, planes)
    
    if cstr is None:
        cstr = interpolate.CubicSpline(
            x=(0, 0.5, 1.0, 2.0, 2.5, 3.0, 3.5, 4.0, 8.0, 255.0),
            y=(0, 0.1, 0.6, 0.9, 1.0, 1.09, 1.15, 1.19, 1.249, 1.5)
        )(sstr)

    if ldmp is None:
        ldmp = sstr + 0.1

    blur_kernel = BlurMatrix.BINOMIAL()
    blur_kernel2 = blur_kernel

    if mode < 0:
        cstr **= 0.8
        blur_kernel2 = box_blur

    mode = abs(mode)

    if mode == 1:
        blurred = median_blur(blur_kernel(func.work_clip))
    elif mode > 1:
        blurred = blur_kernel(median_blur(func.work_clip))
    if mode == 3:
        blurred = median_blur(blurred)

    diff = norm_expr(
        [func.work_clip, blurred],
        'range_size 256 / SCL! x y - SCL@ / D! D@ abs DA! DA@ {lstr} / 1 {pstr} / pow {sstr} * '
        'D@ DA@ 0.001 + / * D@ 2 pow D@ 2 pow {ldmp} + / * SCL@ * neutral +',
        lstr=lstr, pstr=pstr, sstr=sstr, ldmp=ldmp
    )

    sharp = func.work_clip

    if sstr:
        sharp = sharp.std.MergeDiff(diff)

    if cstr:
        diff = norm_expr(diff, 'x neutral - {cstr} * neutral +', cstr=cstr)
        diff = blur_kernel2(diff)
        sharp = sharp.std.MakeDiff(diff)

    if xstr:
        xysharp = norm_expr([sharp, box_blur(sharp)], 'x x y - 9.9 * +')
        rpsharp = repair(xysharp, sharp, 12)
        sharp = rpsharp.std.Merge(sharp, weight=[1 - xstr])

    return func.return_clip(sharp)


def soothe(
    flt: vs.VideoNode, src: vs.VideoNode, spatial_strength: int = 0, temporal_strength: int = 25,
    spatial_radius: int = 1, temporal_radius: int = 1, scenechange: bool = False, planes: PlanesT = 0
) -> vs.VideoNode:
    sharp_diff = src.std.MakeDiff(flt, planes)

    expr = (
        'x neutral - X! y neutral - Y! X@ 0 < Y@ 0 < xor X@ 100 / {strength} * '
        'X@ abs Y@ abs > X@ {strength} * Y@ 100 {strength} - * + 100 / X@ ? ? neutral +'
    )

    if spatial_strength:
        soothe = box_blur(sharp_diff, radius=spatial_radius, planes=planes)
        strength = 100 - abs(max(min(spatial_strength, 100), 0))
        sharp_diff = norm_expr([sharp_diff, soothe], expr, strength=strength, planes=planes)

    if temporal_strength:
        soothe = (
            BlurMatrix.MEAN(temporal_radius, mode=ConvMode.TEMPORAL)
            (sharp_diff, planes=planes, scenechange=scenechange)
        )
        strength = 100 - abs(max(min(temporal_strength, 100), -100))
        sharp_diff = norm_expr([sharp_diff, soothe], expr, strength=strength, planes=planes)

    return src.std.MakeDiff(sharp_diff, planes)



from __future__ import annotations

import operator

from enum import auto
from itertools import accumulate
from math import ceil, exp, log2, pi, sqrt
from typing import Any, Iterable, Literal, Self, Sequence, overload

from vsexprtools import ExprList, ExprOp, ExprToken, ExprVars
from vstools import (
    ConvMode, CustomIntEnum, CustomValueError, KwargsT, Nb, PlanesT, check_variable, core, fallback,
    iterate, shift_clip_multi, to_singleton, vs
)

__all__ = [
    'LimitFilterMode',
    'RemoveGrainMode', 'RemoveGrainModeT',
    'RepairMode', 'RepairModeT',
    'VerticalCleanerMode', 'VerticalCleanerModeT',
    'BlurMatrixBase', 'BlurMatrix'
]


class LimitFilterModeMeta:
    force_expr = True


class LimitFilterMode(LimitFilterModeMeta, CustomIntEnum):
    """Two sources, one filtered"""
    SIMPLE_MIN = auto()
    SIMPLE_MAX = auto()
    """One source, two filtered"""
    SIMPLE2_MIN = auto()
    SIMPLE2_MAX = auto()
    DIFF_MIN = auto()
    DIFF_MAX = auto()
    """One/Two sources, one filtered"""
    CLAMPING = auto()

    @property
    def op(self) -> str:
        return '<' if 'MIN' in self._name_ else '>'

    def __call__(self, force_expr: bool = True) -> LimitFilterMode:
        self.force_expr = force_expr

        return self


class RemoveGrainMode(CustomIntEnum):
    NONE = 0
    MINMAX_AROUND1 = 1
    MINMAX_AROUND2 = 2
    MINMAX_AROUND3 = 3
    MINMAX_MEDIAN = 4
    EDGE_CLIP_STRONG = 5
    EDGE_CLIP_MODERATE = 6
    EDGE_CLIP_MEDIUM = 7
    EDGE_CLIP_LIGHT = 8
    LINE_CLIP_CLOSE = 9
    MIN_SHARP = 10
    BINOMIAL_BLUR = 11
    BOB_TOP_CLOSE = 13
    BOB_BOTTOM_CLOSE = 14
    BOB_TOP_INTER = 15
    BOB_BOTTOM_INTER = 16
    MINMAX_MEDIAN_OPP = 17
    LINE_CLIP_OPP = 18
    BOX_BLUR_NO_CENTER = 19
    BOX_BLUR = 20
    OPP_CLIP_AVG = 21
    OPP_CLIP_AVG_FAST = 22
    EDGE_DEHALO = 23
    EDGE_DEHALO2 = 24
    MIN_SHARP2 = 25
    SMART_RGC = 26
    SMART_RGCL = 27
    SMART_RGCL2 = 28

    def __call__(self, clip: vs.VideoNode, planes: PlanesT = None) -> vs.VideoNode:
        from .rgtools import removegrain
        from .util import norm_rmode_planes
        return removegrain(clip, norm_rmode_planes(clip, self, planes))


RemoveGrainModeT = int | RemoveGrainMode | Sequence[int | RemoveGrainMode]


class RepairMode(CustomIntEnum):
    NONE = 0
    MINMAX_SQUARE1 = 1
    MINMAX_SQUARE2 = 2
    MINMAX_SQUARE3 = 3
    MINMAX_SQUARE4 = 4
    LINE_CLIP_MIN = 5
    LINE_CLIP_LIGHT = 6
    LINE_CLIP_MEDIUM = 7
    LINE_CLIP_STRONG = 8
    LINE_CLIP_CLOSE = 9
    MINMAX_SQUARE_REF_CLOSE = 10
    MINMAX_SQUARE_REF1 = 11
    MINMAX_SQUARE_REF2 = 12
    MINMAX_SQUARE_REF3 = 13
    MINMAX_SQUARE_REF4 = 14
    CLIP_REF_RG5 = 15
    CLIP_REF_RG6 = 16
    CLIP_REF_RG17 = 17
    CLIP_REF_RG18 = 18
    CLIP_REF_RG19 = 19
    CLIP_REF_RG20 = 20
    CLIP_REF_RG21 = 21
    CLIP_REF_RG22 = 22
    CLIP_REF_RG23 = 23
    CLIP_REF_RG24 = 24
    CLIP_REF_RG26 = 26
    CLIP_REF_RG27 = 27
    CLIP_REF_RG28 = 28

    def __call__(self, clip: vs.VideoNode, repairclip: vs.VideoNode, planes: PlanesT = None) -> vs.VideoNode:
        from .rgtools import repair
        from .util import norm_rmode_planes
        return repair(clip, repairclip, norm_rmode_planes(clip, self, planes))


RepairModeT = int | RepairMode | Sequence[int | RepairMode]


class VerticalCleanerMode(CustomIntEnum):
    NONE = 0
    MEDIAN = 1
    PRESERVING = 2

    def __call__(self, clip: vs.VideoNode, planes: PlanesT = None) -> vs.VideoNode:
        from .rgtools import vertical_cleaner
        from .util import norm_rmode_planes
        return vertical_cleaner(clip, norm_rmode_planes(clip, self, planes))


VerticalCleanerModeT = int | VerticalCleanerMode | Sequence[int | VerticalCleanerMode]


class BlurMatrixBase(list[Nb]):
    def __init__(
        self, __iterable: Iterable[Nb], /, mode: ConvMode = ConvMode.SQUARE,
    ) -> None:
        self.mode = mode
        super().__init__(__iterable)  # type: ignore[arg-type]

    def __call__(
        self, clip: vs.VideoNode, planes: PlanesT = None,
        bias: float | None = None, divisor: float | None = None, saturate: bool = True,
        passes: int = 1, expr_kwargs: KwargsT | None = None, **conv_kwargs: Any
    ) -> vs.VideoNode:
        """
        Performs a spatial or temporal convolution.
        It will either calls std.Convolution, std.AverageFrames or ExprOp.convolution
        based on the ConvMode mode picked.

        :param clip:            Clip to process.
        :param planes:          Specifies which planes will be processed.
        :param bias:            Value to add to the final result of the convolution
                                (before clamping the result to the format's range of valid values).
        :param divisor:         Divide the output of the convolution by this value (before adding bias).
                                The default is the sum of the elements of the matrix
        :param saturate:        If True, negative values become 0.
                                If False, absolute values are returned.
        :param passes:          Number of iterations.
        :param expr_kwargs:     A KwargsT of keyword arguments for ExprOp.convolution.__call__ when it is picked.
        :param **conv_kwargs:   Additional keyword arguments for std.Convolution, std.AverageFrames or ExprOp.convolution.

        :return:                Processed clip.
        """
        if len(self) <= 1:
            return clip

        assert (check_variable(clip, self.__call__))

        expr_kwargs = expr_kwargs or KwargsT()

        fp16 = clip.format.sample_type == vs.FLOAT and clip.format.bits_per_sample == 16

        if self.mode.is_spatial:
            # std.Convolution is limited to 25 numbers
            # SQUARE mode is not optimized
            # std.Convolution doesn't support float 16
            if len(self) <= 25 and self.mode != ConvMode.SQUARE and not fp16:
                return iterate(clip, core.std.Convolution, passes, self, bias, divisor, planes, saturate, self.mode)

            return iterate(
                clip, ExprOp.convolution("x", self, bias, fallback(divisor, True), saturate, self.mode, **conv_kwargs),
                passes, planes=planes, **expr_kwargs
            )

        if all([
            not fp16,
            len(self) <= 31,
            not bias,
            saturate,
            (len(conv_kwargs) == 0 or (len(conv_kwargs) == 1 and "scenechange" in conv_kwargs))
        ]):
            return iterate(clip, core.std.AverageFrames, passes, self, divisor, planes=planes, **conv_kwargs)

        return self._averageframes_akarin(clip, planes, bias, divisor, saturate, passes, expr_kwargs, **conv_kwargs)

    def _averageframes_akarin(self, *args: Any, **kwargs: Any) -> vs.VideoNode:
        clip, planes, bias, divisor, saturate, passes, expr_kwargs = args
        conv_kwargs = kwargs

        r = len(self) // 2

        if conv_kwargs.pop("scenechange", False) is False:
            expr_conv = ExprOp.convolution(
                ExprVars(len(self)), self, bias, fallback(divisor, True), saturate, self.mode, **conv_kwargs
            )
            return iterate(
                clip, lambda x: expr_conv(shift_clip_multi(x, (-r, r)), planes=planes, **expr_kwargs), passes
            )

        vars_, = ExprOp.matrix(ExprVars(len(self)), r, self.mode)

        # Conditionnal for backward frames
        condb = list([ExprList(["cond0!", "cond0@", 0])])

        for i, vv in enumerate(accumulate(vars_[:r][::-1], operator.add), 1):
            ww = list[Any]()

            for j, (v, w) in enumerate(zip(vv, self[:r][::-1])):
                ww.append([v, w, ExprOp.DUP, f"div{j}!", ExprOp.MUL])

            condb.append(ExprList([f"cond{i}!", f"cond{i}@", ww, [ExprOp.ADD] * (i - 1)]))

        # Conditionnal for forward frames
        condf = list([ExprList([f"cond{i + 1}!", f"cond{i + 1}@", 0])])

        for ii, vv in enumerate(accumulate(vars_[r + 1:], operator.add), i + 2):
            ww = list[Any]()

            for jj, (v, w) in enumerate(zip(vv, self[r + 1:]), j + 2):
                ww.append([v, w, ExprOp.DUP, f"div{jj}!", ExprOp.MUL])

            condf.append(ExprList([f"cond{ii}!", f"cond{ii}@", ww, [ExprOp.ADD] * (ii - i - 2)]))

        expr = ExprList()

        # Conditionnal for backward frames
        for i, (v, c) in enumerate(zip(vars_[:r][::-1], condb)):
            expr.append(f"{v}._SceneChangeNext", *c)

        expr.append(condb[-1][2:], ExprOp.TERN * r)

        # Conditionnal for forward frames
        for i, (v, c) in enumerate(zip(vars_[r + 1:], condf)):
            expr.append(f"{v}._SceneChangePrev", *c)

        expr.append(condf[-1][2:], ExprOp.TERN * r)

        expr.append(ExprOp.ADD, vars_[r], self[r])

        # Center frame
        weights_cum_b = ExprList(v for v in reversed(list(accumulate(self[:r]))))
        weights_cum_f = ExprList(v for v in reversed(list(accumulate(self[r + 1:][::-1]))))

        for k, ws in zip(range(ii), weights_cum_b + ExprList() + weights_cum_f):
            if k == r:
                continue
            expr.append(f"cond{k}@", ws, ExprOp.MUL)

        expr.extend([ExprOp.ADD] * k)
        expr.append(ExprOp.DUP, f"div{r}!", ExprOp.MUL, ExprOp.ADD)

        if (premultiply := conv_kwargs.get("premultiply", None)):
            expr.append(premultiply, ExprOp.MUL)

        if divisor:
            expr.append(divisor, ExprOp.DIV)
        else:
            # Divisor conditionnal
            for cond, rr in zip([condb, condf], [(0, r), (r + 1, r * 2 + 1)]):
                for n, m in zip(accumulate(([d] for d in range(*rr)), initial=[0]), cond[:-1]):
                    if (n := n[1:]):
                        div = list[str]()
                        for divn in n:
                            div.append(f"div{divn}@")
                    else:
                        div = [str(0)]

                    expr.append(str(m[0])[:5] + "@", div, [ExprOp.ADD] * max(0, len(n) - 1))

                expr.append(*div, f"div{divn + 1}@", ExprOp.ADD * len(div))
                expr.append(ExprOp.TERN * r)

            expr.append(ExprOp.ADD, f"div{r}@", ExprOp.ADD, ExprOp.DIV)

        if bias:
            expr.append(bias, ExprOp.ADD)

        if not saturate:
            expr.append(ExprOp.ABS)

        if (multiply := conv_kwargs.get("multiply", None)):
            expr.append(multiply, ExprOp.MUL)

        if conv_kwargs.get("clamp", False):
            expr.append(ExprOp.clamp(ExprToken.RangeMin, ExprToken.RangeMax))

        return iterate(clip, lambda x: expr(shift_clip_multi(x, (-r, r)), planes=planes, **expr_kwargs), passes)

    def outer(self) -> Self:
        from numpy import outer

        return self.__class__(outer(self, self).flatten().tolist(), self.mode)


class BlurMatrix(CustomIntEnum):
    CIRCLE = 0
    MEAN = 1
    BINOMIAL = 2
    LOG = 3

    @to_singleton.as_property
    class GAUSS:
        def __call__(
            self,
            taps: int | None = None,
            *,
            sigma: float = 0.5,
            mode: ConvMode = ConvMode.HV,
            **kwargs: Any
        ) -> BlurMatrixBase[float]:
            scale_value = kwargs.get("scale_value", 1023)

            if mode == ConvMode.SQUARE:
                scale_value = sqrt(scale_value)

            taps = self.get_taps(sigma, taps)

            if taps < 0:
                raise CustomValueError('Taps must be >= 0!')

            if sigma > 0.0:
                half_pisqrt = 1.0 / sqrt(2.0 * pi) * sigma
                doub_qsigma = 2 * sigma ** 2

                high, *mat = [half_pisqrt * exp(-x ** 2 / doub_qsigma) for x in range(taps + 1)]

                mat = [x * scale_value / high for x in mat]
                mat = [*mat[::-1], scale_value, *mat]
            else:
                mat = [scale_value]

            kernel = BlurMatrixBase(mat, mode)

            if mode == ConvMode.SQUARE:
                kernel = kernel.outer()

            return kernel

        def from_radius(self, radius: int) -> BlurMatrixBase[float]:
            return self(None, sigma=(radius + 1.0) / 3)

        @staticmethod
        def get_taps(sigma: float, taps: int | None = None) -> int:
            if taps is None:
                taps = ceil(abs(sigma) * 8 + 1) // 2

            return taps

    @overload
    def __call__(  # type: ignore[misc]
        self: Literal[BlurMatrix.CIRCLE], taps: int = 1, *, mode: ConvMode = ConvMode.SQUARE
    ) -> BlurMatrixBase[int]:
        ...

    @overload
    def __call__(  # type: ignore[misc]
        self: Literal[BlurMatrix.MEAN], taps: int = 1, *, mode: ConvMode = ConvMode.SQUARE
    ) -> BlurMatrixBase[int]:
        ...

    @overload
    def __call__(  # type: ignore[misc]
        self: Literal[BlurMatrix.BINOMIAL], taps: int = 1, *, mode: ConvMode = ConvMode.HV, **kwargs: Any
    ) -> BlurMatrixBase[int]:
        ...

    @overload
    def __call__(  # type: ignore[misc]
        self: Literal[BlurMatrix.LOG], taps: int = 1, *, strength: float = 100.0, mode: ConvMode = ConvMode.HV
    ) -> BlurMatrixBase[float]:
        ...

    def __call__(self, taps: int = 1, **kwargs: Any) -> Any:
        kernel: BlurMatrixBase[Any]

        match self:
            case BlurMatrix.CIRCLE:
                mode = kwargs.pop("mode", ConvMode.SQUARE)

                matrix = [1 for _ in range(((2 * taps + 1) ** (2 if mode == ConvMode.SQUARE else 1)) - 1)]
                matrix.insert(len(matrix) // 2, 0)

                return BlurMatrixBase[int](matrix, mode)

            case BlurMatrix.MEAN:
                mode = kwargs.pop("mode", ConvMode.SQUARE)

                kernel = BlurMatrixBase[int]([1 for _ in range(((2 * taps + 1)))], mode)

            case BlurMatrix.BINOMIAL:
                mode = kwargs.pop("mode", ConvMode.HV)

                c = 1
                n = taps * 2 + 1

                matrix = list[int]()

                for i in range(1, taps + 2):
                    matrix.append(c)
                    c = c * (n - i) // i

                kernel = BlurMatrixBase(matrix[:-1] + matrix[::-1], mode)

            case BlurMatrix.LOG:
                mode = kwargs.pop("mode", ConvMode.HV)
                strength = kwargs.get("strength", 100)

                strength = max(1e-6, min(log2(3) * strength / 100, log2(3)))

                weight = 0.5 ** strength / ((1 - 0.5 ** strength) * 0.5)

                matrixf = [1.0]

                for _ in range(taps):
                    matrixf.append(matrixf[-1] / weight)

                kernel = BlurMatrixBase([*matrixf[::-1], *matrixf[1:]], mode)

        if mode == ConvMode.SQUARE:
            kernel = kernel.outer()

        return kernel



from __future__ import annotations

from functools import partial
from itertools import count
from typing import Any, Literal, overload

from vsexprtools import ExprOp, ExprVars, complexpr_available, norm_expr
from vskernels import Bilinear, Gaussian
from vstools import (
    ConvMode, CustomValueError, FunctionUtil, OneDimConvModeT, PlanesT, SpatialConvModeT,
    TempConvModeT, check_variable, core, depth, get_depth, join, normalize_planes, normalize_seq, split, to_arr, vs
)

from .enum import BlurMatrix, BlurMatrixBase, LimitFilterMode
from .freqs import MeanMode
from .limit import limit_filter
from .util import normalize_radius

__all__ = [
    'box_blur', 'side_box_blur',
    'gauss_blur',
    'min_blur', 'sbr', 'median_blur',
    'bilateral', 'flux_smooth'
]


def box_blur(
    clip: vs.VideoNode, radius: int | list[int] = 1, passes: int = 1,
    mode: OneDimConvModeT | TempConvModeT = ConvMode.HV, planes: PlanesT = None, **kwargs: Any
) -> vs.VideoNode:
    assert check_variable(clip, box_blur)

    planes = normalize_planes(clip, planes)

    if isinstance(radius, list):
        return normalize_radius(clip, box_blur, radius, planes, passes=passes)

    if not radius:
        return clip

    if mode == ConvMode.TEMPORAL or (clip.format.sample_type == vs.FLOAT and clip.format.bits_per_sample == 16):
        return BlurMatrix.MEAN(radius, mode=mode)(clip, planes, passes=passes, **kwargs)

    box_args = (
        planes,
        radius, 0 if mode == ConvMode.VERTICAL else passes,
        radius, 0 if mode == ConvMode.HORIZONTAL else passes
    )

    if hasattr(core, 'vszip'):
        return clip.vszip.BoxBlur(*box_args)

    if radius > 12:
        return clip.std.BoxBlur(*box_args)

    return BlurMatrix.MEAN(radius, mode=mode)(clip, planes, passes=passes, **kwargs)


def side_box_blur(
    clip: vs.VideoNode, radius: int | list[int] = 1, planes: PlanesT = None,
    inverse: bool = False
) -> vs.VideoNode:
    planes = normalize_planes(clip, planes)

    if isinstance(radius, list):
        return normalize_radius(clip, side_box_blur, radius, planes, inverse=inverse)

    half_kernel = [(1 if i <= 0 else 0) for i in range(-radius, radius + 1)]

    conv_m1 = partial(core.std.Convolution, matrix=half_kernel, planes=planes)
    conv_m2 = partial(core.std.Convolution, matrix=half_kernel[::-1], planes=planes)
    blur_pt = partial(box_blur, planes=planes)

    vrt_filters, hrz_filters = [
        [
            partial(conv_m1, mode=mode), partial(conv_m2, mode=mode),
            partial(blur_pt, hradius=hr, vradius=vr, hpasses=h, vpasses=v)
        ] for h, hr, v, vr, mode in [
            (0, None, 1, radius, ConvMode.VERTICAL), (1, radius, 0, None, ConvMode.HORIZONTAL)
        ]
    ]

    vrt_intermediates = (vrt_flt(clip) for vrt_flt in vrt_filters)
    intermediates = list(
        hrz_flt(vrt_intermediate)
        for i, vrt_intermediate in enumerate(vrt_intermediates)
        for j, hrz_flt in enumerate(hrz_filters) if not i == j == 2
    )

    comp_blur = None if inverse else box_blur(clip, radius, 1, planes=planes)

    if complexpr_available:
        template = '{cum} x - abs {new} x - abs < {cum} {new} ?'

        cum_expr, cumc = '', 'y'
        n_inter = len(intermediates)

        for i, newc, var in zip(count(), ExprVars[2:26], ExprVars[4:26]):
            if i == n_inter - 1:
                break

            cum_expr += template.format(cum=cumc, new=newc)

            if i != n_inter - 2:
                cumc = var.upper()
                cum_expr += f' {cumc}! '
                cumc = f'{cumc}@'

        if comp_blur:
            clips = [clip, *intermediates, comp_blur]
            cum_expr = f'x {cum_expr} - {ExprVars[n_inter + 1]} +'
        else:
            clips = [clip, *intermediates]

        cum = norm_expr(clips, cum_expr, planes, force_akarin='vsrgtools.side_box_blur')
    else:
        cum = intermediates[0]
        for new in intermediates[1:]:
            cum = limit_filter(clip, cum, new, LimitFilterMode.SIMPLE2_MIN, planes)

        if comp_blur:
            cum = clip.std.MakeDiff(cum).std.MergeDiff(comp_blur)

    if comp_blur:
        return box_blur(cum, 1, min(radius // 2, 1))

    return cum


def gauss_blur(
    clip: vs.VideoNode, sigma: float | list[float] = 0.5, taps: int | None = None,
    mode: ConvMode = ConvMode.HV, planes: PlanesT = None,
    **kwargs: Any
) -> vs.VideoNode:
    assert check_variable(clip, gauss_blur)

    planes = normalize_planes(clip, planes)

    if isinstance(sigma, list):
        return normalize_radius(clip, gauss_blur, ('sigma', sigma), planes, mode=mode)

    if ConvMode.VERTICAL in mode:
        sigma = min(sigma, clip.height)

    if ConvMode.HORIZONTAL in mode:
        sigma = min(sigma, clip.width)

    taps = BlurMatrix.GAUSS.get_taps(sigma, taps)

    if hasattr(core, 'resize2') and not mode.is_temporal:
        def _resize2_blur(plane: vs.VideoNode, sigma: float, taps: int) -> vs.VideoNode:
            resize_kwargs = dict[str, Any]()

            # Downscale approximation can be used by specifying _fast=True
            # Has a big speed gain when taps is large
            if kwargs.pop("_fast", False):
                wdown, hdown = plane.width, plane.height

                if ConvMode.VERTICAL in mode:
                    hdown = round(max(round(hdown / sigma), 2) / 2) * 2

                if ConvMode.HORIZONTAL in mode:
                    wdown = round(max(round(wdown / sigma), 2) / 2) * 2

                resize_kwargs.update(width=plane.width, height=plane.height)

                plane = Bilinear.scale(plane, wdown, hdown)
                sigma = 0.8952637851149309
                taps = min(taps, 128)
            else:
                resize_kwargs.update({f'force_{k}': k in mode for k in 'hv'})

            return Gaussian(sigma, taps).scale(plane, **resize_kwargs | kwargs)

        if not {*range(clip.format.num_planes)} - {*planes}:
            return _resize2_blur(clip, sigma, taps)

        return join([
            _resize2_blur(p, sigma, taps) if i in planes else p
            for i, p in enumerate(split(clip))
        ])

    kernel: BlurMatrixBase[float] = BlurMatrix.GAUSS(  # type: ignore
        taps, sigma=sigma, mode=mode, scale_value=1.0 if taps > 12 else 1023
    )

    return kernel(clip, planes, **kwargs)


def min_blur(
    clip: vs.VideoNode, radius: int | list[int] = 1,
    mode: tuple[ConvMode, ConvMode] = (ConvMode.HV, ConvMode.SQUARE), planes: PlanesT = None,
    **kwargs: Any
) -> vs.VideoNode:
    """
    MinBlur by Didée (http://avisynth.nl/index.php/MinBlur)
    Nifty Gauss/Median combination
    """
    assert check_variable(clip, min_blur)

    planes = normalize_planes(clip, planes)

    if isinstance(radius, list):
        return normalize_radius(clip, min_blur, radius, planes)
    
    mode_blur, mode_median = normalize_seq(mode, 2)

    blurred = BlurMatrix.BINOMIAL(radius=radius, mode=mode_blur)(clip, planes=planes, **kwargs)
    median = median_blur(clip, radius, mode_median, planes=planes)

    return MeanMode.MEDIAN([clip, blurred, median], planes=planes)


def sbr(
    clip: vs.VideoNode, radius: int | list[int] = 1,
    mode: ConvMode = ConvMode.HV, planes: PlanesT = None,
    **kwargs: Any
) -> vs.VideoNode:
    assert check_variable(clip, sbr)

    planes = normalize_planes(clip, planes)

    blur_kernel = BlurMatrix.BINOMIAL(radius=radius, mode=mode)

    blurred = blur_kernel(clip, planes=planes, **kwargs)

    diff = clip.std.MakeDiff(blurred, planes=planes)
    blurred_diff = blur_kernel(diff, planes=planes, **kwargs)

    return norm_expr(
        [clip, diff, blurred_diff],
        'y z - D1! y neutral - D2! x D1@ D2@ xor 0 D1@ abs D2@ abs < D1@ D2@ ? ? -',
        planes=planes
    )


@overload
def median_blur(
    clip: vs.VideoNode, radius: int = ..., mode: Literal[ConvMode.TEMPORAL] = ..., planes: PlanesT = ...
) -> vs.VideoNode:
    ...


@overload
def median_blur(
    clip: vs.VideoNode, radius: int | list[int] = ..., mode: SpatialConvModeT = ..., planes: PlanesT = None
) -> vs.VideoNode:
    ...


@overload
def median_blur(
    clip: vs.VideoNode, radius: int | list[int] = ..., mode: ConvMode = ..., planes: PlanesT = None
) -> vs.VideoNode:
    ...


def median_blur(
    clip: vs.VideoNode, radius: int | list[int] = 1, mode: ConvMode = ConvMode.SQUARE, planes: PlanesT = None
) -> vs.VideoNode:
    if mode == ConvMode.TEMPORAL:
        if isinstance(radius, int):
            return clip.zsmooth.TemporalMedian(radius, planes)

        raise CustomValueError("A list of radius isn't supported for ConvMode.TEMPORAL!", median_blur, radius)

    radius = to_arr(radius)

    if (len((rs := set(radius))) == 1 and rs.pop() == 1) and mode == ConvMode.SQUARE:
        return clip.std.Median(planes=planes)

    expr_plane = list[list[str]]()

    for r in radius:
        expr_passes = list[str]()

        for mat in ExprOp.matrix('x', r, mode, [(0, 0)]):
            rb = len(mat) + 1
            st = rb - 1
            sp = rb // 2 - 1
            dp = st - 2

            expr_passes.append(f"{mat} sort{st} swap{sp} min! swap{sp} max! drop{dp} x min@ max@ clip")

        expr_plane.append(expr_passes)

    for e in zip(*expr_plane):
        clip = norm_expr(clip, e, planes, force_akarin=median_blur)

    return clip


def bilateral(
    clip: vs.VideoNode, sigmaS: float | list[float] = 3.0, sigmaR: float | list[float] = 0.02,
    ref: vs.VideoNode | None = None, radius: int | list[int] | None = None,
    device_id: int = 0, num_streams: int | None = None, use_shared_memory: bool = True,
    block_x: int | None = None, block_y: int | None = None, planes: PlanesT = None,
    *, gpu: bool | None = None
) -> vs.VideoNode:
    func = FunctionUtil(clip, bilateral, planes)

    sigmaS, sigmaR = func.norm_seq(sigmaS), func.norm_seq(sigmaR)

    if gpu is not False:
        basic_args, new_args = (sigmaS, sigmaR, radius, device_id), (num_streams, use_shared_memory)

        if hasattr(core, 'bilateralgpu_rtc'):
            return clip.bilateralgpu_rtc.Bilateral(*basic_args, *new_args, block_x, block_y, ref)
        else:
            return clip.bilateralgpu.Bilateral(*basic_args, *new_args, ref)

    if (bits := get_depth(clip)) > 16:
        clip = depth(clip, 16)

    if ref and clip.format != ref.format:
        ref = depth(ref, clip)

    clip = clip.vszip.Bilateral(ref, sigmaS, sigmaR)

    return depth(clip, bits)


def flux_smooth(
    clip: vs.VideoNode, temporal_threshold: float = 7.0, spatial_threshold: float = 0.0,
    scalep: bool = True, planes: PlanesT = None
) -> vs.VideoNode:
    func = FunctionUtil(clip, flux_smooth, planes)

    if spatial_threshold:
        smoothed = func.work_clip.zsmooth.FluxSmoothST(temporal_threshold, spatial_threshold, scalep)
    else:
        smoothed = func.work_clip.zsmooth.FluxSmoothT(temporal_threshold, scalep)

    return func.return_clip(smoothed)
